import os

import grpc

from union.filesystems._middleware import (
    StreamStreamClientAuthInterceptor,
    StreamUnaryClientAuthInterceptor,
    UnaryStreamClientAuthInterceptor,
    UnaryUnaryClientAuthInterceptor,
)

_ENDPOINT = "localhost:8080"
MAX_MESSAGE_LENGTH = 10 * 1024 * 1024


def _get_endpoint() -> str:
    endpoint = os.environ.get("OBJECT_STORE_ENDPOINT")
    if endpoint is not None and endpoint != "":
        return endpoint

    endpoint = os.environ.get("object_store_endpoint")
    if endpoint is not None and endpoint != "":
        return endpoint

    return _ENDPOINT


def _is_remote(endpoint: str) -> bool:
    return endpoint != _ENDPOINT


def _create_channel():
    endpoint = _get_endpoint()

    interceptors = None
    if _is_remote(endpoint):
        unary_unary_auth_interceptor = UnaryUnaryClientAuthInterceptor()
        unary_stream_auth_interceptor = UnaryStreamClientAuthInterceptor()
        stream_unary_auth_interceptor = StreamUnaryClientAuthInterceptor()
        stream_stream_auth_interceptor = StreamStreamClientAuthInterceptor()
        interceptors = [
            unary_unary_auth_interceptor,
            unary_stream_auth_interceptor,
            stream_unary_auth_interceptor,
            stream_stream_auth_interceptor,
        ]

    channel = grpc.aio.insecure_channel(
        endpoint,
        options=[
            ("grpc.max_message_length", MAX_MESSAGE_LENGTH),
            ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
        ],
        interceptors=interceptors,
    )
    return channel



from importlib.metadata import version
from textwrap import dedent

import rich
import rich_click as click
from rich.panel import Panel

from union._config import _get_config_obj


@click.command("info")
@click.pass_context
def info(ctx: click.Context):
    config_obj = _get_config_obj(ctx.obj.get("config_file", None))
    endpoint = config_obj.platform.endpoint

    union_version = version("union")
    flytekit_version = version("flytekit")
    context = dedent(
        f"""
    union is the CLI to interact with Union. Use the CLI to register, create and track task and workflow executions locally and remotely.

    Union Version    : {union_version}
    Flytekit Version : {flytekit_version}
    Union Endpoint   : {endpoint}
    """  # noqa
    )

    rich.print(Panel(context, title="Union CLI Info"))



"""All imports to union in this file should be in the function definition.

This plugin is loaded by flytekit, so any imports to union can lead to circular imports.
"""

from contextlib import suppress
from functools import partial
from typing import Optional

import click
from click import Group, Option
from flytekit.remote import FlyteRemote


class UnionAIPlugin:
    @staticmethod
    def get_remote(
        config: Optional[str],
        project: str,
        domain: str,
        data_upload_location: Optional[str] = None,
    ) -> FlyteRemote:
        from union._config import _get_config_obj
        from union.remote import UnionRemote

        return UnionRemote(
            _get_config_obj(config),
            default_project=project,
            default_domain=domain,
            data_upload_location=data_upload_location,
        )

    @staticmethod
    def configure_pyflyte_cli(main: Group) -> Group:
        """Configure pyflyte's CLI."""

        import union._config
        from union._config import _UNION_CONFIG, _get_config_obj
        from union.cli._create import create
        from union.cli._delete import delete
        from union.cli._get import get
        from union.cli._info import info
        from union.cli._update import update
        from union.ucimage._image_builder import _register_union_image_builder

        def _cli_main_config_callback(ctx, param, value):
            # Set org based on config from `pyflyte --config` for downstream cli
            # commands to use.
            if value is not None:
                _UNION_CONFIG.config = value

            # Only register image builder for serverless
            with suppress(click.ClickException):
                cfg_obj = _get_config_obj(value)
                if union._config._should_enable_image_builder(cfg_obj.platform.endpoint):
                    _register_union_image_builder()

            return value

        for p in main.params:
            if p.name == "config":
                p.callback = _cli_main_config_callback

        # Configure org at the top level:
        def _set_org(ctx, param, value):
            union._config._UNION_CONFIG.org = value

        main.params.append(
            click.Option(
                ["--org"],
                help="Set organization",
                hidden=True,
                callback=_set_org,
                expose_value=False,
            )
        )

        def update_or_create_group(new_group):
            try:
                main_group = main.commands[new_group.name]
                for name, command in new_group.commands.items():
                    main_group.add_command(command, name)
            except KeyError:
                main.add_command(new_group)

        new_groups = [create, delete, update, get]
        for group in new_groups:
            update_or_create_group(group)

        new_commands = [info]
        for command in new_commands:
            main.add_command(command)

        def yield_all_options():
            commands = [main]

            while commands:
                command = commands.pop()
                yield from (p for p in command.params if isinstance(p, Option))

                if isinstance(command, Group):
                    commands.extend(list(command.commands.values()))

        for param in yield_all_options():
            if param.name == "image_config" and param.default is not None and not callable(param.default):
                param.default = lambda: [union._config._get_default_image()]
            elif param.name == "project" and (param.default is not None and not callable(param.default)):
                param.default = partial(union._config._get_default_project, previous_default=param.default)

        return main

    @staticmethod
    def secret_requires_group() -> bool:
        """Return True if secrets require group entry."""
        return False

    @staticmethod
    def get_default_image() -> Optional[str]:
        """Return default image."""

        from union._config import _get_default_image

        return _get_default_image()

    @staticmethod
    def get_auth_success_html(endpoint: str) -> Optional[str]:
        """Get default success html. Return None to use flytekit's default success html."""
        from union._config import _get_auth_success_html

        return _get_auth_success_html(endpoint)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: objectstore/payload.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.common import list_pb2 as common_dot_list__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from union.internal.objectstore import definition_pb2 as objectstore_dot_definition__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19objectstore/payload.proto\x12\x17\x63loudidl.objectstore.v1\x1a\x11\x63ommon/list.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cobjectstore/definition.proto\x1a\x17validate/validate.proto\"\x11\n\x0fMetadataRequest\"\xb9\x01\n\x10MetadataResponse\x12G\n!max_single_part_object_size_bytes\x18\x01 \x01(\x04R\x1cmaxSinglePartObjectSizeBytes\x12-\n\x13min_part_size_bytes\x18\x02 \x01(\x04R\x10minPartSizeBytes\x12-\n\x13max_part_size_bytes\x18\x03 \x01(\x04R\x10maxPartSizeBytes\"\xb4\x01\n\nPutRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12=\n\x08metadata\x18\x02 \x01(\x0b\x32!.cloudidl.objectstore.v1.MetadataR\x08metadata\x12\x37\n\x06object\x18\x03 \x01(\x0b\x32\x1f.cloudidl.objectstore.v1.ObjectR\x06object\"@\n\x0bPutResponse\x12\x1d\n\nsize_bytes\x18\x01 \x01(\x04R\tsizeBytes\x12\x12\n\x04\x65tag\x18\x02 \x01(\tR\x04\x65tag\"<\n\nGetRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\"\xb8\x01\n\x0bGetResponse\x12\x37\n\x06object\x18\x01 \x01(\x0b\x32\x1f.cloudidl.objectstore.v1.ObjectR\x06object\x12=\n\x08metadata\x18\x02 \x01(\x0b\x32!.cloudidl.objectstore.v1.MetadataR\x08metadata\x12\x1d\n\nsize_bytes\x18\x03 \x01(\x04R\tsizeBytes\x12\x12\n\x04\x65tag\x18\x04 \x01(\tR\x04\x65tag\"E\n\x0bListRequest\x12\x36\n\x07request\x18\x01 \x01(\x0b\x32\x1c.cloudidl.common.ListRequestR\x07request\"\x9f\x01\n\x0cListResponse\x12\x30\n\x04keys\x18\x01 \x03(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x04keys\x12\x1d\n\nnext_token\x18\x02 \x01(\tR\tnextToken\x12>\n\x0b\x64irectories\x18\x03 \x03(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x0b\x64irectories\"?\n\rDeleteRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\"\x10\n\x0e\x44\x65leteResponse\"=\n\x0bHeadRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\"\x80\x01\n\x0cHeadResponse\x12=\n\x08metadata\x18\x01 \x01(\x0b\x32!.cloudidl.objectstore.v1.MetadataR\x08metadata\x12\x12\n\x04\x65tag\x18\x02 \x01(\tR\x04\x65tag\x12\x1d\n\nsize_bytes\x18\x03 \x01(\x04R\tsizeBytes\"\xbb\x01\n\x17SuccessfulUploadRequest\x12\x61\n\x0b\x65tags_parts\x18\x01 \x03(\x0b\x32@.cloudidl.objectstore.v1.SuccessfulUploadRequest.EtagsPartsEntryR\netagsParts\x1a=\n\x0f\x45tagsPartsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x05R\x05value:\x02\x38\x01\"\x14\n\x12\x41\x62ortUploadRequest\"\xbe\x02\n\x1fTerminateMultipartUploadRequest\x12-\n\x0coperation_id\x18\x01 \x01(\tB\n\xfa\x42\x07r\x05\x10\x01\x18\xc8\x01R\x0boperationId\x12.\n\x03key\x18\x02 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12_\n\x11successful_upload\x18\x03 \x01(\x0b\x32\x30.cloudidl.objectstore.v1.SuccessfulUploadRequestH\x00R\x10successfulUpload\x12P\n\x0c\x61\x62ort_upload\x18\x04 \x01(\x0b\x32+.cloudidl.objectstore.v1.AbortUploadRequestH\x00R\x0b\x61\x62ortUploadB\t\n\x07request\"f\n TerminateMultipartUploadResponse\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12\x12\n\x04\x65tag\x18\x02 \x01(\tR\x04\x65tag\"\x8c\x01\n\x1bStartMultipartUploadRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12=\n\x08metadata\x18\x02 \x01(\x0b\x32!.cloudidl.objectstore.v1.MetadataR\x08metadata\"\x88\x01\n\x1cStartMultipartUploadResponse\x12-\n\x0coperation_id\x18\x01 \x01(\tB\n\xfa\x42\x07r\x05\x10\x01\x18\xc8\x01R\x0boperationId\x12\x39\n\nexpires_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\texpiresAt\"\x8a\x02\n\x11UploadPartRequest\x12-\n\x0coperation_id\x18\x01 \x01(\tB\n\xfa\x42\x07r\x05\x10\x01\x18\xc8\x01R\x0boperationId\x12.\n\x03key\x18\x02 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12+\n\x0bpart_number\x18\x03 \x01(\x05\x42\n\xfa\x42\x07\x1a\x05\x18\x90N \x01R\npartNumber\x12\x37\n\x06object\x18\x04 \x01(\x0b\x32\x1f.cloudidl.objectstore.v1.ObjectR\x06object\x12\x30\n\x0e\x63ontent_length\x18\x05 \x01(\x03\x42\t\x18\x01\xfa\x42\x04\"\x02 \x00R\rcontentLength\"(\n\x12UploadPartResponse\x12\x12\n\x04\x65tag\x18\x01 \x01(\tR\x04\x65tag\"_\n%ListInProgressMultipartUploadsRequest\x12\x36\n\x07request\x18\x01 \x01(\x0b\x32\x1c.cloudidl.common.ListRequestR\x07request\"p\n\x0fMultipartUpload\x12-\n\x0coperation_id\x18\x01 \x01(\tB\n\xfa\x42\x07r\x05\x10\x01\x18\xc8\x01R\x0boperationId\x12.\n\x03key\x18\x02 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\"r\n&ListInProgressMultipartUploadsResponse\x12H\n\noperations\x18\x01 \x03(\x0b\x32(.cloudidl.objectstore.v1.MultipartUploadR\noperations\"\x93\x01\n\x13\x44ownloadPartRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12$\n\tstart_pos\x18\x02 \x01(\x03\x42\x07\xfa\x42\x04\"\x02(\x00R\x08startPos\x12&\n\nsize_bytes\x18\x03 \x01(\x03\x42\x07\xfa\x42\x04\"\x02(\x00R\tsizeBytes\"O\n\x14\x44ownloadPartResponse\x12\x37\n\x06object\x18\x01 \x01(\x0b\x32\x1f.cloudidl.objectstore.v1.ObjectR\x06object\"\x91\x01\n\x0b\x43opyRequest\x12;\n\nsource_key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\tsourceKey\x12\x45\n\x0f\x64\x65stination_key\x18\x02 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x0e\x64\x65stinationKey\"\x0e\n\x0c\x43opyResponse\"\xa3\x02\n\x0ePresignRequest\x12.\n\x03key\x18\x01 \x01(\x0b\x32\x1c.cloudidl.objectstore.v1.KeyR\x03key\x12\x38\n\nexpires_in\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\texpiresIn\x12M\n\x0bput_request\x18\x03 \x01(\x0b\x32*.cloudidl.objectstore.v1.PresignPutRequestH\x00R\nputRequest\x12M\n\x0bget_request\x18\x04 \x01(\x0b\x32*.cloudidl.objectstore.v1.PresignGetRequestH\x00R\ngetRequestB\t\n\x07request\"0\n\x0fPresignResponse\x12\x1d\n\nsigned_url\x18\x01 \x01(\tR\tsignedUrlB\xdb\x01\n\x1b\x63om.cloudidl.objectstore.v1B\x0cPayloadProtoH\x02P\x01Z.github.com/unionai/cloud/gen/pb-go/objectstore\xa2\x02\x03\x43OX\xaa\x02\x17\x43loudidl.Objectstore.V1\xca\x02\x17\x43loudidl\\Objectstore\\V1\xe2\x02#Cloudidl\\Objectstore\\V1\\GPBMetadata\xea\x02\x19\x43loudidl::Objectstore::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'objectstore.payload_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033com.cloudidl.objectstore.v1B\014PayloadProtoH\002P\001Z.github.com/unionai/cloud/gen/pb-go/objectstore\242\002\003COX\252\002\027Cloudidl.Objectstore.V1\312\002\027Cloudidl\\Objectstore\\V1\342\002#Cloudidl\\Objectstore\\V1\\GPBMetadata\352\002\031Cloudidl::Objectstore::V1'
  _SUCCESSFULUPLOADREQUEST_ETAGSPARTSENTRY._options = None
  _SUCCESSFULUPLOADREQUEST_ETAGSPARTSENTRY._serialized_options = b'8\001'
  _TERMINATEMULTIPARTUPLOADREQUEST.fields_by_name['operation_id']._options = None
  _TERMINATEMULTIPARTUPLOADREQUEST.fields_by_name['operation_id']._serialized_options = b'\372B\007r\005\020\001\030\310\001'
  _STARTMULTIPARTUPLOADRESPONSE.fields_by_name['operation_id']._options = None
  _STARTMULTIPARTUPLOADRESPONSE.fields_by_name['operation_id']._serialized_options = b'\372B\007r\005\020\001\030\310\001'
  _UPLOADPARTREQUEST.fields_by_name['operation_id']._options = None
  _UPLOADPARTREQUEST.fields_by_name['operation_id']._serialized_options = b'\372B\007r\005\020\001\030\310\001'
  _UPLOADPARTREQUEST.fields_by_name['part_number']._options = None
  _UPLOADPARTREQUEST.fields_by_name['part_number']._serialized_options = b'\372B\007\032\005\030\220N \001'
  _UPLOADPARTREQUEST.fields_by_name['content_length']._options = None
  _UPLOADPARTREQUEST.fields_by_name['content_length']._serialized_options = b'\030\001\372B\004\"\002 \000'
  _MULTIPARTUPLOAD.fields_by_name['operation_id']._options = None
  _MULTIPARTUPLOAD.fields_by_name['operation_id']._serialized_options = b'\372B\007r\005\020\001\030\310\001'
  _DOWNLOADPARTREQUEST.fields_by_name['start_pos']._options = None
  _DOWNLOADPARTREQUEST.fields_by_name['start_pos']._serialized_options = b'\372B\004\"\002(\000'
  _DOWNLOADPARTREQUEST.fields_by_name['size_bytes']._options = None
  _DOWNLOADPARTREQUEST.fields_by_name['size_bytes']._serialized_options = b'\372B\004\"\002(\000'
  _globals['_METADATAREQUEST']._serialized_start=193
  _globals['_METADATAREQUEST']._serialized_end=210
  _globals['_METADATARESPONSE']._serialized_start=213
  _globals['_METADATARESPONSE']._serialized_end=398
  _globals['_PUTREQUEST']._serialized_start=401
  _globals['_PUTREQUEST']._serialized_end=581
  _globals['_PUTRESPONSE']._serialized_start=583
  _globals['_PUTRESPONSE']._serialized_end=647
  _globals['_GETREQUEST']._serialized_start=649
  _globals['_GETREQUEST']._serialized_end=709
  _globals['_GETRESPONSE']._serialized_start=712
  _globals['_GETRESPONSE']._serialized_end=896
  _globals['_LISTREQUEST']._serialized_start=898
  _globals['_LISTREQUEST']._serialized_end=967
  _globals['_LISTRESPONSE']._serialized_start=970
  _globals['_LISTRESPONSE']._serialized_end=1129
  _globals['_DELETEREQUEST']._serialized_start=1131
  _globals['_DELETEREQUEST']._serialized_end=1194
  _globals['_DELETERESPONSE']._serialized_start=1196
  _globals['_DELETERESPONSE']._serialized_end=1212
  _globals['_HEADREQUEST']._serialized_start=1214
  _globals['_HEADREQUEST']._serialized_end=1275
  _globals['_HEADRESPONSE']._serialized_start=1278
  _globals['_HEADRESPONSE']._serialized_end=1406
  _globals['_SUCCESSFULUPLOADREQUEST']._serialized_start=1409
  _globals['_SUCCESSFULUPLOADREQUEST']._serialized_end=1596
  _globals['_SUCCESSFULUPLOADREQUEST_ETAGSPARTSENTRY']._serialized_start=1535
  _globals['_SUCCESSFULUPLOADREQUEST_ETAGSPARTSENTRY']._serialized_end=1596
  _globals['_ABORTUPLOADREQUEST']._serialized_start=1598
  _globals['_ABORTUPLOADREQUEST']._serialized_end=1618
  _globals['_TERMINATEMULTIPARTUPLOADREQUEST']._serialized_start=1621
  _globals['_TERMINATEMULTIPARTUPLOADREQUEST']._serialized_end=1939
  _globals['_TERMINATEMULTIPARTUPLOADRESPONSE']._serialized_start=1941
  _globals['_TERMINATEMULTIPARTUPLOADRESPONSE']._serialized_end=2043
  _globals['_STARTMULTIPARTUPLOADREQUEST']._serialized_start=2046
  _globals['_STARTMULTIPARTUPLOADREQUEST']._serialized_end=2186
  _globals['_STARTMULTIPARTUPLOADRESPONSE']._serialized_start=2189
  _globals['_STARTMULTIPARTUPLOADRESPONSE']._serialized_end=2325
  _globals['_UPLOADPARTREQUEST']._serialized_start=2328
  _globals['_UPLOADPARTREQUEST']._serialized_end=2594
  _globals['_UPLOADPARTRESPONSE']._serialized_start=2596
  _globals['_UPLOADPARTRESPONSE']._serialized_end=2636
  _globals['_LISTINPROGRESSMULTIPARTUPLOADSREQUEST']._serialized_start=2638
  _globals['_LISTINPROGRESSMULTIPARTUPLOADSREQUEST']._serialized_end=2733
  _globals['_MULTIPARTUPLOAD']._serialized_start=2735
  _globals['_MULTIPARTUPLOAD']._serialized_end=2847
  _globals['_LISTINPROGRESSMULTIPARTUPLOADSRESPONSE']._serialized_start=2849
  _globals['_LISTINPROGRESSMULTIPARTUPLOADSRESPONSE']._serialized_end=2963
  _globals['_DOWNLOADPARTREQUEST']._serialized_start=2966
  _globals['_DOWNLOADPARTREQUEST']._serialized_end=3113
  _globals['_DOWNLOADPARTRESPONSE']._serialized_start=3115
  _globals['_DOWNLOADPARTRESPONSE']._serialized_end=3194
  _globals['_COPYREQUEST']._serialized_start=3197
  _globals['_COPYREQUEST']._serialized_end=3342
  _globals['_COPYRESPONSE']._serialized_start=3344
  _globals['_COPYRESPONSE']._serialized_end=3358
  _globals['_PRESIGNREQUEST']._serialized_start=3361
  _globals['_PRESIGNREQUEST']._serialized_end=3652
  _globals['_PRESIGNRESPONSE']._serialized_start=3654
  _globals['_PRESIGNRESPONSE']._serialized_end=3702
# @@protoc_insertion_point(module_scope)



from union.internal.common import list_pb2 as _list_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.objectstore import definition_pb2 as _definition_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MetadataRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class MetadataResponse(_message.Message):
    __slots__ = ["max_single_part_object_size_bytes", "min_part_size_bytes", "max_part_size_bytes"]
    MAX_SINGLE_PART_OBJECT_SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    MIN_PART_SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    MAX_PART_SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    max_single_part_object_size_bytes: int
    min_part_size_bytes: int
    max_part_size_bytes: int
    def __init__(self, max_single_part_object_size_bytes: _Optional[int] = ..., min_part_size_bytes: _Optional[int] = ..., max_part_size_bytes: _Optional[int] = ...) -> None: ...

class PutRequest(_message.Message):
    __slots__ = ["key", "metadata", "object"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    OBJECT_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    metadata: _definition_pb2.Metadata
    object: _definition_pb2.Object
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., metadata: _Optional[_Union[_definition_pb2.Metadata, _Mapping]] = ..., object: _Optional[_Union[_definition_pb2.Object, _Mapping]] = ...) -> None: ...

class PutResponse(_message.Message):
    __slots__ = ["size_bytes", "etag"]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    size_bytes: int
    etag: str
    def __init__(self, size_bytes: _Optional[int] = ..., etag: _Optional[str] = ...) -> None: ...

class GetRequest(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ...) -> None: ...

class GetResponse(_message.Message):
    __slots__ = ["object", "metadata", "size_bytes", "etag"]
    OBJECT_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    object: _definition_pb2.Object
    metadata: _definition_pb2.Metadata
    size_bytes: int
    etag: str
    def __init__(self, object: _Optional[_Union[_definition_pb2.Object, _Mapping]] = ..., metadata: _Optional[_Union[_definition_pb2.Metadata, _Mapping]] = ..., size_bytes: _Optional[int] = ..., etag: _Optional[str] = ...) -> None: ...

class ListRequest(_message.Message):
    __slots__ = ["request"]
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    request: _list_pb2.ListRequest
    def __init__(self, request: _Optional[_Union[_list_pb2.ListRequest, _Mapping]] = ...) -> None: ...

class ListResponse(_message.Message):
    __slots__ = ["keys", "next_token", "directories"]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    NEXT_TOKEN_FIELD_NUMBER: _ClassVar[int]
    DIRECTORIES_FIELD_NUMBER: _ClassVar[int]
    keys: _containers.RepeatedCompositeFieldContainer[_definition_pb2.Key]
    next_token: str
    directories: _containers.RepeatedCompositeFieldContainer[_definition_pb2.Key]
    def __init__(self, keys: _Optional[_Iterable[_Union[_definition_pb2.Key, _Mapping]]] = ..., next_token: _Optional[str] = ..., directories: _Optional[_Iterable[_Union[_definition_pb2.Key, _Mapping]]] = ...) -> None: ...

class DeleteRequest(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ...) -> None: ...

class DeleteResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class HeadRequest(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ...) -> None: ...

class HeadResponse(_message.Message):
    __slots__ = ["metadata", "etag", "size_bytes"]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    metadata: _definition_pb2.Metadata
    etag: str
    size_bytes: int
    def __init__(self, metadata: _Optional[_Union[_definition_pb2.Metadata, _Mapping]] = ..., etag: _Optional[str] = ..., size_bytes: _Optional[int] = ...) -> None: ...

class SuccessfulUploadRequest(_message.Message):
    __slots__ = ["etags_parts"]
    class EtagsPartsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    ETAGS_PARTS_FIELD_NUMBER: _ClassVar[int]
    etags_parts: _containers.ScalarMap[str, int]
    def __init__(self, etags_parts: _Optional[_Mapping[str, int]] = ...) -> None: ...

class AbortUploadRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class TerminateMultipartUploadRequest(_message.Message):
    __slots__ = ["operation_id", "key", "successful_upload", "abort_upload"]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SUCCESSFUL_UPLOAD_FIELD_NUMBER: _ClassVar[int]
    ABORT_UPLOAD_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    key: _definition_pb2.Key
    successful_upload: SuccessfulUploadRequest
    abort_upload: AbortUploadRequest
    def __init__(self, operation_id: _Optional[str] = ..., key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., successful_upload: _Optional[_Union[SuccessfulUploadRequest, _Mapping]] = ..., abort_upload: _Optional[_Union[AbortUploadRequest, _Mapping]] = ...) -> None: ...

class TerminateMultipartUploadResponse(_message.Message):
    __slots__ = ["key", "etag"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    etag: str
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., etag: _Optional[str] = ...) -> None: ...

class StartMultipartUploadRequest(_message.Message):
    __slots__ = ["key", "metadata"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    metadata: _definition_pb2.Metadata
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., metadata: _Optional[_Union[_definition_pb2.Metadata, _Mapping]] = ...) -> None: ...

class StartMultipartUploadResponse(_message.Message):
    __slots__ = ["operation_id", "expires_at"]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    EXPIRES_AT_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    expires_at: _timestamp_pb2.Timestamp
    def __init__(self, operation_id: _Optional[str] = ..., expires_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class UploadPartRequest(_message.Message):
    __slots__ = ["operation_id", "key", "part_number", "object", "content_length"]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    PART_NUMBER_FIELD_NUMBER: _ClassVar[int]
    OBJECT_FIELD_NUMBER: _ClassVar[int]
    CONTENT_LENGTH_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    key: _definition_pb2.Key
    part_number: int
    object: _definition_pb2.Object
    content_length: int
    def __init__(self, operation_id: _Optional[str] = ..., key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., part_number: _Optional[int] = ..., object: _Optional[_Union[_definition_pb2.Object, _Mapping]] = ..., content_length: _Optional[int] = ...) -> None: ...

class UploadPartResponse(_message.Message):
    __slots__ = ["etag"]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    etag: str
    def __init__(self, etag: _Optional[str] = ...) -> None: ...

class ListInProgressMultipartUploadsRequest(_message.Message):
    __slots__ = ["request"]
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    request: _list_pb2.ListRequest
    def __init__(self, request: _Optional[_Union[_list_pb2.ListRequest, _Mapping]] = ...) -> None: ...

class MultipartUpload(_message.Message):
    __slots__ = ["operation_id", "key"]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    key: _definition_pb2.Key
    def __init__(self, operation_id: _Optional[str] = ..., key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ...) -> None: ...

class ListInProgressMultipartUploadsResponse(_message.Message):
    __slots__ = ["operations"]
    OPERATIONS_FIELD_NUMBER: _ClassVar[int]
    operations: _containers.RepeatedCompositeFieldContainer[MultipartUpload]
    def __init__(self, operations: _Optional[_Iterable[_Union[MultipartUpload, _Mapping]]] = ...) -> None: ...

class DownloadPartRequest(_message.Message):
    __slots__ = ["key", "start_pos", "size_bytes"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    START_POS_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    start_pos: int
    size_bytes: int
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., start_pos: _Optional[int] = ..., size_bytes: _Optional[int] = ...) -> None: ...

class DownloadPartResponse(_message.Message):
    __slots__ = ["object"]
    OBJECT_FIELD_NUMBER: _ClassVar[int]
    object: _definition_pb2.Object
    def __init__(self, object: _Optional[_Union[_definition_pb2.Object, _Mapping]] = ...) -> None: ...

class CopyRequest(_message.Message):
    __slots__ = ["source_key", "destination_key"]
    SOURCE_KEY_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_KEY_FIELD_NUMBER: _ClassVar[int]
    source_key: _definition_pb2.Key
    destination_key: _definition_pb2.Key
    def __init__(self, source_key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., destination_key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ...) -> None: ...

class CopyResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class PresignRequest(_message.Message):
    __slots__ = ["key", "expires_in", "put_request", "get_request"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    EXPIRES_IN_FIELD_NUMBER: _ClassVar[int]
    PUT_REQUEST_FIELD_NUMBER: _ClassVar[int]
    GET_REQUEST_FIELD_NUMBER: _ClassVar[int]
    key: _definition_pb2.Key
    expires_in: _duration_pb2.Duration
    put_request: _definition_pb2.PresignPutRequest
    get_request: _definition_pb2.PresignGetRequest
    def __init__(self, key: _Optional[_Union[_definition_pb2.Key, _Mapping]] = ..., expires_in: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., put_request: _Optional[_Union[_definition_pb2.PresignPutRequest, _Mapping]] = ..., get_request: _Optional[_Union[_definition_pb2.PresignGetRequest, _Mapping]] = ...) -> None: ...

class PresignResponse(_message.Message):
    __slots__ = ["signed_url"]
    SIGNED_URL_FIELD_NUMBER: _ClassVar[int]
    signed_url: str
    def __init__(self, signed_url: _Optional[str] = ...) -> None: ...



import asyncio
import functools
import inspect

from fsspec.asyn import AsyncFileSystem, async_methods, private


def sync(loop, func, *args, timeout=None, **kwargs):
    """
    Make loop run coroutine until it returns. Runs in other thread

    Examples
    --------
    >>> sync(fsspec.asyn.get_loop(), func, *args,
             timeout=timeout, **kwargs)
    """
    timeout = timeout if timeout else None  # convert 0 or 0.0 to None
    # NB: if the loop is not running *yet*, it is OK to submit work
    # and we will wait for it
    if loop is None or loop.is_closed():
        raise RuntimeError("Loop is not running")
    coro = func(*args, **kwargs)
    return loop.run_until_complete(coro)


def sync_wrapper(func, obj=None):
    """Given a function, make so can be called in async or blocking contexts

    Leave obj=None if defining within a class. Pass the instance if attaching
    as an attribute of the instance.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = obj or args[0]
        return sync(self.loop, func, *args, **kwargs)

    return wrapper


def mirror_sync_methods(obj):
    """Populate sync and async methods for obj

    For each method will create a sync version if the name refers to an async method
    (coroutine) and there is no override in the child class; will create an async
    method for the corresponding sync method if there is no implementation.

    Uses the methods specified in
    - async_methods: the set that an implementation is expected to provide
    - default_async_methods: that can be derived from their sync version in
      AbstractFileSystem
    - AsyncFileSystem: async-specific default coroutines
    """
    from fsspec import AbstractFileSystem

    for method in async_methods + dir(AsyncFileSystem):
        if not method.startswith("_"):
            continue
        smethod = method[1:]
        if private.match(method):
            isco = inspect.iscoroutinefunction(getattr(obj, method, None))
            unsync = getattr(getattr(obj, smethod, False), "__func__", None)
            is_default = unsync is getattr(AbstractFileSystem, smethod, "")
            if isco and is_default:
                mth = sync_wrapper(getattr(obj, method), obj=obj)
                setattr(obj, smethod, mth)
                if not mth.__doc__:
                    mth.__doc__ = getattr(getattr(AbstractFileSystem, smethod, None), "__doc__", "")


async def if_coro(result):
    if asyncio.iscoroutine(result):  # or inspect.iscoroutine,... and so on
        return await result
    else:
        return result



"""All imports to unionai in this file should be in the function definition.

This plugin is loaded by flytekit, so any imports to unionai can lead to circular imports.
"""

import json
import re
import shutil
import tempfile
import warnings
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from importlib.metadata import version
from pathlib import Path
from typing import Optional, Tuple

import click
from flytekit import Labels
from flytekit.core.constants import REQUIREMENTS_FILE_NAME
from flytekit.image_spec.image_spec import _F_IMG_ID, ImageBuildEngine, ImageSpec, ImageSpecBuilder
from flytekit.models.core import execution as core_execution_models
from flytekit.models.filters import ValueIn
from flytekit.remote import FlyteRemote, FlyteWorkflowExecution
from flytekit.remote.remote import MOST_RECENT_FIRST
from flytekit.tools.ignore import DockerIgnore, GitIgnore, IgnoreGroup, StandardIgnore
from flytekit.tools.translator import Options

RECOVER_EXECUTION_PAGE_LIMIT = 1_000
UNIONAI_IMAGE_NAME_KEY = "unionai_image_name"

_PACKAGE_NAME_RE = re.compile(r"^[\w-]+")


def _is_union(package: str) -> bool:
    """Return True if `package` is union or unionai."""
    m = _PACKAGE_NAME_RE.match(package)
    if not m:
        return False
    name = m.group()
    return name in ("union", "unionai")


def get_unionai_for_pypi():
    union_version = version("union")
    if "dev" in union_version:
        return "union"
    else:
        return f"union=={union_version}"


@lru_cache
def _get_remote() -> FlyteRemote:
    from union._config import _should_enable_image_builder
    from union.configuration._plugin import UnionAIPlugin

    remote = UnionAIPlugin.get_remote(
        config=None,
        project="default",
        domain="development",
    )
    if not _should_enable_image_builder(remote.config.platform.endpoint):
        raise ValueError(
            "The UnionAI image builder requires a Serverless endpoint. "
            f"Your current endpoint is {remote.config.platform.endpoint}"
        )
    return remote


@lru_cache
def _get_fully_qualified_image_name(remote, execution):
    return execution.outputs["fully_qualified_image"]


def _build(spec: Path, context: Optional[Path], target_image: str) -> str:
    """Build image using UnionAI."""

    remote = _get_remote()
    start = datetime.now(timezone.utc)

    execution = _get_latest_image_build_execution(remote, target_image)
    if execution is None:
        context_url = "" if context is None else remote.upload_file(context)[1]
        spec_url = remote.upload_file(spec)[1]
        entity = remote.fetch_task(project="system", domain="production", name="build-image")

        execution = remote.execute(
            entity,
            project="system",
            domain="development",
            inputs={"spec": spec_url, "context": context_url, "target_image": target_image},
            options=Options(
                # Replace ':' with '_' since flyte does not allow ':' in the label value
                labels=Labels(values={UNIONAI_IMAGE_NAME_KEY: target_image.replace(":", "_")}),
            ),
        )
        click.secho("👍 Build submitted!", bold=True, fg="yellow")

    console_url = remote.generate_console_url(execution)

    click.secho(
        "⏳ Waiting for build to finish at: " + click.style(console_url, fg="cyan"),
        bold=True,
    )
    execution = remote.wait(execution, poll_interval=timedelta(seconds=1))

    elapsed = str(datetime.now(timezone.utc) - start).split(".")[0]

    if execution.closure.phase == core_execution_models.WorkflowExecutionPhase.SUCCEEDED:
        click.secho(f"✅ Build completed in {elapsed}!", bold=True, fg="green")
    else:
        error_msg = execution.error.message
        raise click.ClickException(
            f"❌ Build failed in {elapsed} at {click.style(console_url, fg='cyan')} with error:\n\n{error_msg}"
        )

    return _get_fully_qualified_image_name(remote, execution)


class UCImageSpecBuilder(ImageSpecBuilder):
    """ImageSpec builder for UnionAI."""

    _SUPPORTED_IMAGE_SPEC_PARAMETERS = {
        "name",
        "builder",
        "python_version",
        "source_root",
        "env",
        "packages",
        "requirements",
        "apt_packages",
        "cuda",
        "cudnn",
        "platform",
        "pip_index",
        "pip_extra_index_url",
        "commands",
    }

    def build_image(self, image_spec: ImageSpec):
        """Build image using UnionAI."""
        image_name = image_spec.image_name()

        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            spec_path, archive_path = self._validate_configuration(image_spec, tmp_path, image_name)
            return _build(spec_path, archive_path, image_name)

    def should_build(self, image_spec: ImageSpec) -> bool:
        """Check whether the image should be built."""
        image_name = image_spec.image_name()
        remote = _get_remote()
        execution = _get_latest_image_build_execution(remote, image_name)
        if execution is None:
            click.secho(f"Image cr.union.ai/{image_name} not found.", fg="blue")
            click.secho("🐳 Build not found, submitting a new build...", bold=True, fg="blue")
            return True
        elif (
            execution.closure.phase == core_execution_models.WorkflowExecutionPhase.QUEUED
            or execution.closure.phase == core_execution_models.WorkflowExecutionPhase.RUNNING
        ):
            click.secho(f"Image cr.union.ai/{image_name} not found.", fg="blue")
            click.secho("🐳 Build found, but not finished, waiting...", bold=True, fg="blue")
            return True
        else:
            click.secho(f"Image cr.union.ai/{image_name} found. Skip building.", fg="blue")
            # make sure the fully qualified image name is cached in the image builder
            ImageBuildEngine._IMAGE_NAME_TO_REAL_NAME[image_name] = _get_fully_qualified_image_name(remote, execution)
            return False

    def _validate_configuration(
        self, image_spec: ImageSpec, tmp_path: Path, image_name: str
    ) -> Tuple[Path, Optional[Path]]:
        """Validate and write configuration for builder."""
        unsupported_parameters = [
            name
            for name, value in vars(image_spec).items()
            if value is not None and name not in self._SUPPORTED_IMAGE_SPEC_PARAMETERS and not name.startswith("_")
        ]
        if unsupported_parameters:
            msg = f"The following parameters are unsupported and ignored: {unsupported_parameters}"
            warnings.warn(msg, UserWarning)

        # Hardcoded for now since our base image only supports 3.11
        supported_python_version = "3.11"
        if image_spec.python_version is not None and not str(image_spec.python_version).startswith(
            supported_python_version
        ):
            raise ValueError(
                f"The union image builder only supports Python {supported_python_version}, please set "
                f'python_version="{supported_python_version}"'
            )

        spec = {"python_version": supported_python_version}
        # Transform image spec into a spec we expect
        if image_spec.apt_packages:
            spec["apt_packages"] = image_spec.apt_packages
        if image_spec.commands:
            spec["commands"] = image_spec.commands
        if image_spec.cuda or image_spec.cudnn:
            spec["enable_gpu"] = True

        env = image_spec.env or {}
        env = {**{_F_IMG_ID: image_name}, **env}
        if env:
            spec["env"] = env

        packages = []
        if image_spec.packages:
            packages.extend(image_spec.packages)

        spec["python_packages"] = packages

        pip_extra_index_urls = []
        if image_spec.pip_index:
            pip_extra_index_urls.append(image_spec.pip_index)

        if image_spec.pip_extra_index_url:
            pip_extra_index_urls.extend(image_spec.pip_extra_index_url)

        spec["pip_extra_index_urls"] = pip_extra_index_urls

        context_path = tmp_path / "build.uc-image-builder"
        context_path.mkdir(exist_ok=True)

        _all_packages = []
        _all_packages.extend(packages)

        if image_spec.requirements:
            shutil.copy2(image_spec.requirements, context_path / REQUIREMENTS_FILE_NAME)
            spec["python_requirements_files"] = [REQUIREMENTS_FILE_NAME]

            with (context_path / REQUIREMENTS_FILE_NAME).open() as f:
                _all_packages.extend([line.strip() for line in f.readlines()])

        # if union is not specified, then add it here.
        if not any(_is_union(p) for p in _all_packages):
            packages.append(get_unionai_for_pypi())

        if image_spec.source_root:
            # Easter egg
            # Load in additional packages before installing pip/apt packages
            vendor_path = Path(image_spec.source_root) / ".vendor"
            if vendor_path.is_dir():
                spec["dist_dirpath"] = ".vendor"

            ignore = IgnoreGroup(image_spec.source_root, [GitIgnore, DockerIgnore, StandardIgnore])
            shutil.copytree(
                image_spec.source_root,
                context_path,
                ignore=shutil.ignore_patterns(*ignore.list_ignored()),
                dirs_exist_ok=True,
            )

        if any(context_path.iterdir()):
            archive_path = Path(shutil.make_archive(tmp_path / "context", "xztar", context_path))
        else:
            archive_path = None

        spec_path = tmp_path / "spec.json"
        with spec_path.open("w") as f:
            json.dump(spec, f)

        return (spec_path, archive_path)


def _get_latest_image_build_execution(remote: FlyteRemote, image_name: str) -> Optional[FlyteWorkflowExecution]:
    """
    Get the latest image build execution for the given image name.
    Returns None if no execution is found or the execution is not in QUEUED, RUNNING, or SUCCEEDED phase.
    """
    # Replace ':' with '_' since flyte does not allow ':' in the label value
    target_image = image_name.replace(":", "_")
    executions, _ = remote.client.list_executions_paginated(
        project="system",
        domain="development",
        limit=1,
        filters=[
            ValueIn("execution_tag.key", [UNIONAI_IMAGE_NAME_KEY]),
            ValueIn("execution_tag.value", [target_image]),
            ValueIn("phase", ["QUEUED", "RUNNING", "SUCCEEDED"]),
        ],
        sort_by=MOST_RECENT_FIRST,
    )

    if len(executions) == 0:
        return None

    return remote.sync_execution(FlyteWorkflowExecution.promote_from_model(executions[0]))


def _register_union_image_builder():
    ImageBuildEngine.register("unionai", UCImageSpecBuilder(), priority=10)
    ImageBuildEngine.register("union", UCImageSpecBuilder(), priority=10)



from flytekit.clis.sdk_in_container.pyflyte import main as pyflyte_main

from union._config import _UNION_CONFIG, _get_union_config_env_var
from union._usage import _configure_tracking_cli


def main(*args, **kwargs):
    """Main CLI entry point for directly calling `union`"""
    _UNION_CONFIG.config = _get_union_config_env_var()
    _UNION_CONFIG.is_direct_union_cli_call = True

    main = _configure_tracking_cli(pyflyte_main)
    return main(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: identity/app_definition.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from union.internal.identity import enums_pb2 as identity_dot_enums__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1didentity/app_definition.proto\x12\x11\x63loudidl.identity\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x14identity/enums.proto\"\xbb\x06\n\x03\x41pp\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x1b\n\tclient_id\x18\x02 \x01(\tR\x08\x63lientId\x12I\n\x13\x63lient_id_issued_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x63lientIdIssuedAt\x12#\n\rredirect_uris\x18\x04 \x03(\tR\x0credirectUris\x12G\n\x0eresponse_types\x18\x05 \x03(\x0e\x32 .cloudidl.identity.ResponseTypesR\rresponseTypes\x12>\n\x0bgrant_types\x18\x06 \x03(\x0e\x32\x1d.cloudidl.identity.GrantTypesR\ngrantTypes\x12M\n\x10\x61pplication_type\x18\x07 \x01(\x0e\x32\".cloudidl.identity.ApplicationTypeR\x0f\x61pplicationType\x12\x1a\n\x08\x63ontacts\x18\x08 \x03(\tR\x08\x63ontacts\x12\x1f\n\x0b\x63lient_name\x18\t \x01(\tR\nclientName\x12\x19\n\x08logo_uri\x18\n \x01(\tR\x07logoUri\x12\x1d\n\nclient_uri\x18\x0b \x01(\tR\tclientUri\x12\x1d\n\npolicy_uri\x18\x0c \x01(\tR\tpolicyUri\x12\x17\n\x07tos_uri\x18\r \x01(\tR\x06tosUri\x12\x19\n\x08jwks_uri\x18\x0e \x01(\tR\x07jwksUri\x12g\n\x1atoken_endpoint_auth_method\x18\x0f \x01(\x0e\x32*.cloudidl.identity.TokenEndpointAuthMethodR\x17tokenEndpointAuthMethod\x12#\n\rclient_secret\x18\x10 \x01(\tR\x0c\x63lientSecret\x12S\n\x18\x63lient_secret_expires_at\x18\x11 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15\x63lientSecretExpiresAtB\xbf\x01\n\x15\x63om.cloudidl.identityB\x12\x41ppDefinitionProtoH\x02P\x01Z+github.com/unionai/cloud/gen/pb-go/identity\xa2\x02\x03\x43IX\xaa\x02\x11\x43loudidl.Identity\xca\x02\x11\x43loudidl\\Identity\xe2\x02\x1d\x43loudidl\\Identity\\GPBMetadata\xea\x02\x12\x43loudidl::Identityb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'identity.app_definition_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\025com.cloudidl.identityB\022AppDefinitionProtoH\002P\001Z+github.com/unionai/cloud/gen/pb-go/identity\242\002\003CIX\252\002\021Cloudidl.Identity\312\002\021Cloudidl\\Identity\342\002\035Cloudidl\\Identity\\GPBMetadata\352\002\022Cloudidl::Identity'
  _globals['_APP']._serialized_start=108
  _globals['_APP']._serialized_end=935
# @@protoc_insertion_point(module_scope)



import base64
import importlib
import os
import sys
import textwrap
from dataclasses import dataclass
from functools import lru_cache, partial
from importlib.metadata import version
from os import getenv
from pathlib import Path
from string import Template
from typing import List, Optional, Union

import click
import grpc
from flyteidl.service.identity_pb2 import UserInfoRequest
from flyteidl.service.identity_pb2_grpc import IdentityServiceStub
from flytekit.clients.auth_helper import get_authenticated_channel
from flytekit.configuration import AuthType, Config, PlatformConfig, get_config_file
from flytekit.configuration.default_images import DefaultImages

_DEFAULT_GCP_SERVERLESS_ENDPOINT: str = "serverless-1.us-east-2.s.union.ai"
_SERVERLESS_ENDPOINTS = {
    "serverless-1.us-east-2.s.union.ai",
    "serverless-gcp.cloud-staging.union.ai",
    "utt-srv-staging-1.cloud-staging.union.ai",
    "serverless-preview.canary.unionai.cloud",
    "utt-srv-canary-1.canary.unionai.cloud",
}

_UNION_CONFIG_ENV_VARS: List[str] = ["UNION_CONFIG", "UNIONAI_CONFIG"]
_UNION_SERVERLESS_ENDPOINT_ENV_VAR: str = "UNION_SERVERLESS_ENDPOINT"
_UNION_API_KEY_ENV_VARS: List[str] = [
    "UNION_API_KEY",
    "UNION_SERVERLESS_API_KEY",
    "UNIONAI_SERVERLESS_API_KEY",
]
_UNION_ENABLE_IMAGE_BUILDER_ENV_VAR: str = "UNION_ENABLE_REGISTER_IMAGE_BUILDER"
_UNION_DEFAULT_CONFIG_PATHS: List[Path] = [
    Path.home() / ".union" / "config.yaml",
    Path.home() / ".uctl" / "config.yaml",
]
_UNION_DEFAULT_CONFIG_PATH: Path = _UNION_DEFAULT_CONFIG_PATHS[0]


@dataclass
class _UnionConfig:
    serverless_endpoint: str = getenv(
        _UNION_SERVERLESS_ENDPOINT_ENV_VAR,
        _DEFAULT_GCP_SERVERLESS_ENDPOINT,
    )
    org: Optional[str] = None
    config: Optional[str] = None
    is_direct_union_cli_call: bool = False


_UNION_CONFIG = _UnionConfig()


def _is_serverless_endpoint(endpoint: str) -> str:
    """Check if endpoint is serverless."""
    return endpoint in (endpoint, f"dns:///{endpoint}") and endpoint in _SERVERLESS_ENDPOINTS


def _should_enable_image_builder(endpoint: str) -> bool:
    """Enable image builder when either holds:

    - `endpoint` is a serverless endpoint
    - UNIONAI_ENABLE_REGISTER_IMAGE_BUILDER=1

    """
    is_serverless = _is_serverless_endpoint(endpoint)
    enable_with_env = os.getenv(_UNION_ENABLE_IMAGE_BUILDER_ENV_VAR, "0") == "1"
    return is_serverless or enable_with_env


@dataclass
class AppClientCredentials:
    endpoint: str
    client_id: str
    client_secret: str
    org: str


def _encode_app_client_credentials(app_credentials: AppClientCredentials) -> str:
    """Encode app_credentials with base64."""
    data = (
        f"{app_credentials.endpoint}:{app_credentials.client_id}:{app_credentials.client_secret}:{app_credentials.org}"
    )
    return base64.b64encode(data.encode("utf-8")).decode("utf-8")


def _decode_app_client_credentials(encoded_str: str) -> AppClientCredentials:
    """Decode encoded base64 string into app credentials."""
    endpoint, client_id, client_secret, org = base64.b64decode(encoded_str.encode("utf-8")).decode("utf-8").split(":")
    return AppClientCredentials(endpoint=endpoint, client_id=client_id, client_secret=client_secret, org=org)


@lru_cache(maxsize=1)
def _unionmeta_byoc_is_installed() -> bool:
    """Check if unionmeta_byoc is installed."""
    try:
        importlib.import_module("unionmeta_byoc")
        return True
    except ModuleNotFoundError:
        return False


def _clean_endpoint(endpoint: str) -> str:
    """Clean endpoint."""
    prefixes = ("dns:///", "http://", "https://")

    for prefix in prefixes:
        if endpoint.startswith(prefix):
            n_prefix = len(prefix)
            endpoint = endpoint[n_prefix:]
            break

    return endpoint.rstrip("/")


def _get_endpoint_for_login(endpoint: Optional[str]) -> str:
    """Get endpoint for login."""
    if _unionmeta_byoc_is_installed():
        # byoc
        if endpoint is None:
            msg = "--host must be passed in for login"
            raise click.ClickException(msg)
        return _clean_endpoint(endpoint)
    else:
        if endpoint is not None:
            return _clean_endpoint(endpoint)
        return _UNION_CONFIG.serverless_endpoint


def _get_env_var(env_vars: List[str]) -> Optional[str]:
    """Get environment variable from a list of env_vars. env_vars that are first"""
    for env_var in env_vars:
        if env_var in os.environ:
            return os.environ[env_var]
    return None


def _get_config_path(config_paths: List[Path] = _UNION_DEFAULT_CONFIG_PATHS) -> Optional[Path]:
    """Get configuration path."""
    for path in config_paths:
        if path.exists():
            return path
    return None


_get_union_config_env_var = partial(_get_env_var, env_vars=_UNION_CONFIG_ENV_VARS)
_get_union_api_env_var = partial(_get_env_var, env_vars=_UNION_API_KEY_ENV_VARS)


def _check_unionmeta_byoc(config: Optional[str]):
    """Check defaults for `pip install union[byoc]`

    - `--config` must be passed in
    - `UNION_CONFIG` must be set.
    """
    if not _unionmeta_byoc_is_installed():
        return

    if config is not None or _UNION_CONFIG.config is not None:
        # Config is passed in directly
        return

    union_env_var = _get_union_config_env_var()
    config_path = _get_config_path()

    if _UNION_CONFIG.is_direct_union_cli_call and union_env_var is None and config_path is None:
        msg = (
            "When using the union CLI, please pass in --config, set UNION_CONFIG or "
            "add a config to ~/.union/config.yaml"
        )

        raise click.ClickException(msg)


def _check_yaml_config_is_empty(config_file: Union[str, os.PathLike]):
    with open(config_file, "r") as f:
        contents = f.read().strip()

    if not contents:
        raise click.ClickException(
            f"Unable to load yaml at {config_file} because it is empty. "
            "Delete file or pass a config that is not empty"
        )


def _decode_union_api_key(serverless_api_value: str) -> Config:
    """Decode union serverless api key and return Config."""
    config = Config.auto()
    try:
        app_credentials = _decode_app_client_credentials(serverless_api_value)
    except Exception as e:
        raise ValueError("Unable to read UNION_API_KEY") from e

    if _UNION_CONFIG.org is None and app_credentials.org != "":
        _UNION_CONFIG.org = app_credentials.org

    return config.with_params(
        platform=PlatformConfig(
            endpoint=app_credentials.endpoint,
            insecure=False,
            auth_mode=AuthType.CLIENTSECRET,
            client_id=app_credentials.client_id,
            client_credentials_secret=app_credentials.client_secret,
        )
    )


def _get_config_obj(config: Optional[str] = None, default_to_union_semantics: bool = False) -> Config:
    """Get Config object.

    If `config` is not None, then it will be used as the Config file.

    If the `union` CLI is called directly or `default_to_union_semantics=True`, then the config is set
    in the following order:
        1. `UNION_SERVERLESS_API_KEY` environment variable
        2. `UNION_CONFIG` environment variable
        3. `UNION_SERVERLESS_ENDPOINT` environment variable
        4. ~/.union/config.yaml if it exists

    If `pyflyte` CLI is called and `flytekit`'s `get_config_file` did not return a `ConfigFile`,
    then serverless it the default endpoint.
    """
    if config is not None:
        return Config.auto(config)

    elif _UNION_CONFIG.is_direct_union_cli_call or default_to_union_semantics:
        _check_unionmeta_byoc(config)

        # CLI ran with `union` -> Config is _UNION_CONFIG.config and the default is serverless
        if (serverless_api_value := _get_union_api_env_var()) is not None and serverless_api_value != "":
            return _decode_union_api_key(serverless_api_value)
        elif _UNION_CONFIG.config is not None:
            return Config.auto(_UNION_CONFIG.config)
        elif getenv(_UNION_SERVERLESS_ENDPOINT_ENV_VAR) is not None:
            # If UNION_SERVERLESS_ENDPOINT is set in `_UnionConfig`, use it instead
            return Config.for_endpoint(endpoint=_UNION_CONFIG.serverless_endpoint)
        elif (config_path := _get_config_path()) is not None:
            # Check if config is empty:
            _check_yaml_config_is_empty(config_path)
            # Read config from file system
            return Config.auto(str(config_path.absolute()))
        else:
            # Got to serverless
            return Config.for_endpoint(endpoint=_UNION_CONFIG.serverless_endpoint)
    else:
        # CLI ran with `pyflyte`, we use mostly the same semantics as `pyflyte`. The difference is:
        # - byoc => Go to localhost
        # - else => Go to serverless
        cfg_file = get_config_file(_UNION_CONFIG.config)

        if cfg_file is not None:
            return Config.auto(cfg_file)

        # At this point, cfg_file is None. For byoc, we keep the flytekit behavior and go to localhost
        if _unionmeta_byoc_is_installed():
            return Config.auto(cfg_file)

        # Without byoc, then go to serverless
        return Config.for_endpoint(endpoint=_UNION_CONFIG.serverless_endpoint)


def _get_organization(platform_config: PlatformConfig, channel: Optional[grpc.Channel] = None) -> str:
    """Get organization based on endpoint."""
    if _UNION_CONFIG.org is not None:
        return _UNION_CONFIG.org
    elif _is_serverless_endpoint(platform_config.endpoint):
        org = _get_user_handle(platform_config, channel)
        _UNION_CONFIG.org = org
        return org
    else:
        # Managed+ users, the org is not required for requests and we set it ""
        # to replicate default flytekit behavior.
        return ""


def _get_user_handle(platform_config: PlatformConfig, channel: Optional[grpc.Channel] = None) -> str:
    """Get user_handle for PlatformConfig."""
    if channel is None:
        channel = get_authenticated_channel(platform_config)

    client = IdentityServiceStub(channel)
    user_info = client.UserInfo(UserInfoRequest())
    user_handle = user_info.additional_claims.fields["userhandle"]
    return user_handle.string_value


def _get_default_image() -> str:
    """Get default image version."""
    cfg_obj = _get_config_obj()

    # TODO: This is only temporary to support GCP endpoints. When the union images are public,
    # we will always use union images
    endpoint = cfg_obj.platform.endpoint
    if _is_serverless_endpoint(endpoint):
        major, minor = sys.version_info.major, sys.version_info.minor
        union_version = version("union")
        if "dev" in union_version:
            suffix = "latest"
        else:
            suffix = union_version

        return f"cr.union.ai/v1/unionai/union:py{major}.{minor}-{suffix}"

    return DefaultImages().find_image_for()


def _write_config_to_path(endpoint: str, auth_type: str, config_path: Path = _UNION_DEFAULT_CONFIG_PATH):
    """Write config to config directory."""
    config_dir = config_path.parent
    config_dir.mkdir(exist_ok=True, parents=True)

    config_template = Template(
        textwrap.dedent(
            """\
    admin:
      endpoint: $endpoint
      insecure: false
      authType: $auth_type
    logger:
      show-source: true
      level: 0
    union:
      connection:
        host: $endpoint
        insecure: false
      auth:
        type: $auth_type
    """
        )
    )
    config_path.write_text(config_template.substitute(endpoint=f"dns:///{endpoint}", auth_type=auth_type))
    return config_path


def _get_default_project(previous_default: str):
    cfg_obj = _get_config_obj()
    if _is_serverless_endpoint(cfg_obj.platform.endpoint):
        return "default"
    return previous_default


def _get_auth_success_html(endpoint: str) -> str:
    """Get default success html. Return None to use flytekit's default success html."""
    if endpoint.endswith("union.ai") or endpoint.endswith("unionai.cloud"):
        SUCCESS_HTML = textwrap.dedent(
            f"""
        <html>
        <head>
            <title>OAuth2 Authentication to Union Successful</title>
        </head>
        <body style="background:white;font-family:Arial">
            <div style="position: absolute;top:40%;left:50%;transform: translate(-50%, -50%);text-align:center;">
                <div style="margin:auto">
                    <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 300 65" fill="currentColor"
                        style="color:#fdb51e;width:360px;">
                        <title>Union.ai</title>
                        <path d="M32,64.8C14.4,64.8,0,51.5,0,34V3.6h17.6v41.3c0,1.9,1.1,3,3,3h23c1.9,0,3-1.1,3-3V3.6H64V34
                        C64,51.5,49.6,64.8,32,64.8z M69.9,30.9v30.4h17.6V20c0-1.9,1.1-3,3-3h23c1.9,0,3,1.1,3,3v41.3H134V30.9c0-17.5-14.4-30.8-32.1-30.8
                        S69.9,13.5,69.9,30.9z M236,30.9v30.4h17.6V20c0-1.9,1.1-3,3-3h23c1.9,0,3,1.1,3,3v41.3H300V30.9c0-17.5-14.4-30.8-32-30.8
                        S236,13.5,236,30.9L236,30.9z M230.1,32.4c0,18.2-14.2,32.5-32.2,32.5s-32-14.3-32-32.5s14-32.1,32-32.1S230.1,14.3,230.1,32.4
                        L230.1,32.4z M213.5,20.2c0-1.9-1.1-3-3-3h-24.8c-1.9,0-3,1.1-3,3v24.5c0,1.9,1.1,3,3,3h24.8c1.9,0,3-1.1,3-3V20.2z M158.9,3.6
                        h-17.6v57.8h17.6V3.6z"></path>
                    </svg>
                    <h2>You've successfully authenticated to Union!</h2>
                    <p style="font-size:20px;">Return to your terminal for next steps</p>
                </div>
            </div>
        </body>
        </html>
        """  # noqa
        )
        return SUCCESS_HTML
    return None



from __future__ import annotations

import datetime as _datetime
import hashlib
import logging
from dataclasses import asdict, dataclass
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from flytekit.configuration import SerializationSettings
from flytekit.core import launch_plan as _annotated_launchplan
from flytekit.core import workflow as _annotated_workflow
from flytekit.core.base_task import TaskResolverMixin
from flytekit.core.python_auto_container import get_registerable_container_image
from flytekit.core.python_function_task import PythonFunctionTask
from flytekit.core.resources import Resources
from flytekit.core.task import TaskPlugins
from flytekit.core.task import task as flytekit_task_decorator
from flytekit.extras.accelerators import BaseAccelerator
from flytekit.image_spec import ImageSpec
from flytekit.models.documentation import Documentation
from flytekit.models.security import Secret

logger = logging.getLogger(__name__)


@dataclass
class ActorEnvironment:
    """
    ActorEnvironment class.

    :param name: The name of the actor. This is used in conjunction with the
        project, domain, and version to uniquely identify the actor.
    :param container_image: The container image to use for the task. Set to
        default image if none provided.
    :param backlog_length: The number of tasks to keep in the worker queue on
        the backend. Setting `backlog_length` ensures that the worker
        executing actor tasks immediately executes the next task after
        completing the previous one instead of waiting for the scheduler to
        complete other operations before scheduling the next task.
    :param parallelism: The number of tasks that can execute in parallel,
        per worker.
    :param replica_count: The number of workers to provision that are able to
        accept tasks.
    :param ttl_seconds: How long to keep the Actor alive while no tasks are
        being run.
    :param environment: Environment variables as key, value pairs in a Python
        dictionary.
    :param requests: Compute resource requests per task.
    :param limits: Compute resource limits.
    :param accelerator: The accelerator device to use for the task.
    :param secret_requests: Keys (ideally descriptive) that can identify the
        secrets supplied at runtime.
    """

    # These fields here are part of the actor config.
    name: str
    container_image: Optional[Union[str, ImageSpec]] = None
    backlog_length: Optional[int] = None
    parallelism: int = 1
    replica_count: int = 1
    ttl_seconds: int = 100
    environment: Optional[Dict[str, str]] = None
    requests: Optional[Resources] = None
    limits: Optional[Resources] = None
    accelerator: Optional[BaseAccelerator] = None
    secret_requests: Optional[List[Secret]] = None

    def _task(
        self,
        _task_function=None,
        cache: bool = False,
        cache_serialize: bool = False,
        cache_version: str = "",
        retries: int = 0,
        timeout: Union[_datetime.timedelta, int] = 0,
        node_dependency_hints: Optional[
            Iterable[
                Union[
                    PythonFunctionTask,
                    _annotated_launchplan.LaunchPlan,
                    _annotated_workflow.WorkflowBase,
                ]
            ]
        ] = None,
        task_resolver: Optional[TaskResolverMixin] = None,
        docs: Optional[Documentation] = None,
        execution_mode: PythonFunctionTask.ExecutionBehavior = PythonFunctionTask.ExecutionBehavior.DEFAULT,
        **kwargs,
    ):
        wrapper = partial(
            flytekit_task_decorator,
            task_config=self,
            cache=cache,
            cache_serialize=cache_serialize,
            cache_version=cache_version,
            retries=retries,
            timeout=timeout,
            node_dependency_hints=node_dependency_hints,
            task_resolver=task_resolver,
            docs=docs,
            container_image=self.container_image,
            environment=self.environment,
            requests=self.requests,
            limits=self.limits,
            accelerator=self.accelerator,
            secret_requests=self.secret_requests,
            execution_mode=execution_mode,
            **kwargs,
        )

        if _task_function:
            return wrapper(_task_function=_task_function)
        return wrapper

    @property
    def task(self):
        return self._task

    # Keeping this private for now to test
    @property
    def dynamic(self):
        return partial(self._task, execution_mode=PythonFunctionTask.ExecutionBehavior.DYNAMIC)

    @property
    def version(self) -> str:
        fields = []
        for k, v in asdict(self).items():
            if isinstance(v, BaseAccelerator):
                fields.append(str(v.to_flyte_idl()))
            else:
                fields.append(str(v))
        all_fields = "".join(fields)
        hash_object = hashlib.md5(all_fields.encode("utf-8"))
        hex_digest = hash_object.hexdigest()
        logger.debug(f"Computing actor hash with string {all_fields} to {hex_digest[:15]}")
        return hex_digest[:15]


class ActorTask(PythonFunctionTask[ActorEnvironment]):
    _ACTOR_TASK_TYPE = "fast-task"

    def __init__(self, task_config: ActorEnvironment, task_function: Callable, **kwargs):
        super(ActorTask, self).__init__(
            task_config=task_config,
            task_type=self._ACTOR_TASK_TYPE,
            task_function=task_function,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Optional[Dict[str, Any]]:
        """
        Serialize the `ActorTask` config into a dict.

        :param settings: Current serialization settings
        :return: Dictionary representation of the dask task config.
        """
        return {
            "name": self.task_config.name,
            "version": self.task_config.version,
            "type": self._ACTOR_TASK_TYPE,
            "spec": {
                "container_image": get_registerable_container_image(self.container_image, settings.image_config),
                "backlog_length": self.task_config.backlog_length,
                "parallelism": self.task_config.parallelism,
                "replica_count": self.task_config.replica_count,
                "ttl_seconds": self.task_config.ttl_seconds,
            },
        }


TaskPlugins.register_pythontask_plugin(ActorEnvironment, ActorTask)



import typing
from logging import getLogger
from typing import AsyncIterable, Callable

import grpc

DEFAULT_KUBERNETES_TOKEN_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/token"

logger = getLogger(__name__)


# See https://github.com/grpc/grpc/issues/34537 and proposed fix https://github.com/grpc/grpc/pull/36487
# Unfortunately we can't use one interceptor for unary and streaming endpoints
class ServiceAccountMetadataInterceptor:
    def __init__(self):
        self._service_account_token = self._read_service_account_token()

    def _read_service_account_token(self) -> typing.Optional[None]:
        try:
            with open(DEFAULT_KUBERNETES_TOKEN_PATH, "r") as token_file:
                token = token_file.read()
            return token
        except Exception as e:
            logger.debug(f"Error reading service account token: {e}, not adding to request")
            return None

    def _inject_default_metadata(self, client_call_details: grpc.aio.ClientCallDetails) -> grpc.aio.ClientCallDetails:
        if self._service_account_token is None:
            return client_call_details

        old_metadata = client_call_details.metadata or []
        new_metadata = old_metadata + [("authorization", f"Bearer {self._service_account_token}")]

        new_details = grpc.aio.ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=new_metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )
        return new_details


class UnaryUnaryClientAuthInterceptor(grpc.aio.UnaryUnaryClientInterceptor, ServiceAccountMetadataInterceptor):
    async def intercept_unary_unary(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: typing.Any
    ) -> typing.Any:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request)


class UnaryStreamClientAuthInterceptor(grpc.aio.UnaryStreamClientInterceptor, ServiceAccountMetadataInterceptor):
    async def intercept_unary_stream(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: any
    ) -> AsyncIterable:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request)


class StreamUnaryClientAuthInterceptor(grpc.aio.StreamUnaryClientInterceptor, ServiceAccountMetadataInterceptor):
    async def intercept_stream_unary(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: any,
    ) -> grpc.aio.StreamUnaryCall:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)


class StreamStreamClientAuthInterceptor(grpc.aio.StreamStreamClientInterceptor, ServiceAccountMetadataInterceptor):
    async def intercept_stream_stream(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: typing.AsyncIterator[any],
    ) -> grpc.aio.Call:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)



"""Module for fsspec implementations."""

from union._logging import _init_global_loggers
from union.filesystems._unionfs import AsyncUnionFS
from union.filesystems._unionmetafs import AsyncUnionMetaFS

__all__ = ["AsyncUnionFS", "AsyncUnionMetaFS"]

# Configure loggers when our fsspec protocol is called, but union is not imported yet
_init_global_loggers()



import inspect
import os
import platform
import subprocess
import sys
import tarfile
from dataclasses import dataclass
from multiprocessing import Process
from pathlib import Path
from shutil import which
from string import Template
from subprocess import run
from textwrap import dedent
from time import sleep, time
from typing import List, Optional, Union

import flytekit
import fsspec
from flytekit import ImageSpec, PythonFunctionTask, Resources, current_context
from flytekit.configuration import AuthType, SerializationSettings
from flytekit.core.base_task import Task
from flytekit.core.context_manager import FlyteContextManager
from flytekit.core.utils import ClassDecorator
from flytekit.extend import TaskResolverMixin
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin

VSCODE_TYPE_VALUE = "vscode"
# Config keys to store in task template
TASK_FUNCTION_SOURCE_PATH = "TASK_FUNCTION_SOURCE_PATH"

CODE_SERVER_CLI_NAME = "code-server"
DEFAULT_PORT = 8080
DOWNLOAD_DIR = Path.home() / ".code-server"
WORKSPACE_DIR = Path("/workspace")
HEARTBEAT_PATH = Path.home() / ".local" / "share" / "code-server" / "heartbeat"
CODE_SERVER_VERSION = "4.23.1"
CODE_SERVER_DOWNLOAD_URL = (
    "https://github.com/coder/code-server/releases/download/v{version}/code-server-{version}-linux-{arch}.tar.gz"
)
CODE_SERVER_CLI_NAME_TEMPLATE = "code-server-{version}-linux-{arch}"
HEARTBEAT_CHECK_SECONDS = 60
# 20 minutes is the default up time
DEFAULT_UP_SECONDS = 20 * 60

ROOT_PATH = Path.home()
BASHRC_PATH = Path.home() / ".bashrc"
UNION_GITHUB_TOKEN_ENV = "UNION_WORKSPACE_GITHUB_TOKEN"
UNION_SERVERLESS_API_KEY_ENV = "UNION_WORKSPACE_SERVERLESS_API_KEY"
GIT_CONFIG_PATH = Path.home() / ".gitconfig"


GIT_CONFIG_TEMPLATE = Template(
    """\
$USER_INFO

[credential "https://github.com"]
    helper =
    helper = !/usr/bin/gh auth git-credential
[credential "https://gist.github.com"]
    helper =
    helper = !/usr/bin/gh auth git-credential

[url "https://$GITHUB_TOKEN@github.com/"]
    insteadOf = https://github.com/
"""
)

_DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE = """\
name: my-workspace
# Make sure that the project and domain exists
project: flytesnacks
domain: development
container_image: ghcr.io/unionai/workspace:0.0.9

resources:
    cpu: "2"
    mem: "4Gi"
    gpu: null

workspace_root: ~/workspace
on_startup: null
metadata: null
"""

_DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC = """\
name: my-workspace
project: default
domain: development
container_image:
    builder: "union"

resources:
    cpu: "2"
    mem: "4Gi"
    gpu: null

on_startup: null
metadata: null
"""


class ImageSpecConfig(DataClassYAMLMixin, DataClassJSONMixin, ImageSpec):
    def __hash__(self):
        return hash(self.to_dict().__str__())


class ResourceConfig(DataClassYAMLMixin, Resources):
    pass


@dataclass
class WorkspaceConfig(DataClassYAMLMixin, DataClassJSONMixin):
    name: str
    project: str
    domain: str
    container_image: Optional[Union[ImageSpecConfig, str]]
    resources: Optional[ResourceConfig] = None
    workspace_root: str = "/workspace"
    on_startup: Optional[str] = None

    def __post_init__(self):
        # Quick hack to always build image
        if not isinstance(self.container_image, str) and os.getenv("UNION_DEV"):
            from union._testing import imagespec_with_local_union
            from union.ucimage._image_builder import _register_union_image_builder

            _register_union_image_builder()
            image_spec = imagespec_with_local_union()
            self.container_image = ImageSpecConfig(
                name="workspace",
                builder="union",
                packages=image_spec.packages + ["keyrings.alt"],
                source_root=image_spec.source_root,
                apt_packages=["git", "sudo", "wget", "vim"],
                commands=[
                    "wget https://github.com/cli/cli/releases/download/v2.49.0/gh_2.49.0_linux_amd64.deb "
                    "-O /tmp/gh.deb",
                    "apt install /tmp/gh.deb",
                    "wget https://github.com/coder/code-server/releases/download/v4.23.1/code-server_4.23.1_amd64.deb "
                    "-O /tmp/code-server.deb",
                    "apt install /tmp/code-server.deb",
                ],
            )


def configure_python_path():
    python_bin = os.path.dirname(sys.executable)
    with BASHRC_PATH.open("a") as f:
        f.write(os.linesep)
        f.write(f"export PATH={python_bin}:$PATH")
        f.write(os.linesep)
        f.write(r'export PS1="\[\e[0;34m\]\w\[\e[m\]\$ "')


def configure_union_remote(host: str):
    from union._config import _write_config_to_path

    _write_config_to_path(host, AuthType.DEVICEFLOW.value)


def configure_git(repo: str, workspace_dir: Path, git_name: str, git_email: str) -> str:
    """Configures workspace built with ImageSpec on Union's hosted image builder."""

    ctx = current_context()
    secrets = ctx.secrets

    with BASHRC_PATH.open("a") as f:
        github_token = secrets.get(key=UNION_GITHUB_TOKEN_ENV)
        f.write(f"export GITHUB_TOKEN={github_token}")

    user_info = dedent(
        f"""\
    [user]
        name = {git_name}
        email = {git_email}
    """
    )
    git_config = GIT_CONFIG_TEMPLATE.substitute(
        USER_INFO=user_info,
        GITHUB_TOKEN=github_token,
    )

    GIT_CONFIG_PATH.write_text(git_config)

    root_dir = workspace_dir

    if repo != "":
        workspace_dir.mkdir(exist_ok=True)
        subprocess.run(["/usr/bin/git", "clone", repo], cwd=workspace_dir, text=True)
        for item in workspace_dir.iterdir():
            if item.is_dir():
                root_dir = item
                break
    return root_dir


class vscode(ClassDecorator):
    """Union specific VSCode extension."""

    def __init__(self, task_function: Optional[callable] = None):
        super().__init__(task_function=task_function)

    def execute(self, *args, **kwargs):
        ctx = FlyteContextManager.current_context()
        ctx.user_space_params.builder().add_attr(
            TASK_FUNCTION_SOURCE_PATH, inspect.getsourcefile(self.task_function)
        ).build()

        if ctx.execution_state.is_local_execution():
            return

        self.task_function(*args, **kwargs)

    def get_extra_config(self):
        return {self.LINK_TYPE_KEY: VSCODE_TYPE_VALUE, self.PORT_KEY: f"{DEFAULT_PORT}"}


@vscode
def workspace(config_content: str, host: str):
    ROOT_PATH.mkdir(exist_ok=True)
    configure_python_path()
    configure_union_remote(host)

    workspace_config = WorkspaceConfig.from_yaml(config_content)
    workspace_dir = Path(workspace_config.workspace_root).expanduser()

    workspace_dir.mkdir(exist_ok=True)
    config_path = workspace_dir / "config.yaml"
    config_path.write_text(config_content)

    on_startup = workspace_config.on_startup
    if on_startup is not None:
        commands = on_startup.split()
        subprocess.run(commands, cwd=workspace_dir, text=True)

    start_vscode_service(workspace_dir)


class WorkspaceResolver(TaskResolverMixin):
    @property
    def location(self) -> str:
        return "union.workspace._vscode.resolver"

    @property
    def name(self) -> str:
        return "union.workspace.vscode"

    def get_task_for_workspace(
        self,
        config: WorkspaceConfig,
    ) -> PythonFunctionTask:
        return PythonFunctionTask(
            task_config=None,
            task_function=workspace,
            task_type="workspace",
            task_resolver=self,
            requests=config.resources,
            container_image=config.container_image,
        )

    def load_task(self, loader_args: List[str]) -> PythonFunctionTask:
        return PythonFunctionTask(
            task_config=None,
            task_function=workspace,
            task_type="workspace",
            task_resolver=self,
        )

    def loader_args(self, settings: SerializationSettings, task: PythonFunctionTask) -> List[str]:
        return ["workspace"]

    def get_all_tasks(self) -> List[Task]:
        raise NotImplementedError


resolver = WorkspaceResolver()


def download_http(url: str, local_dest_path: Path) -> Path:
    """Download URL to `download_dir`. Returns Path of downloaded file."""
    logger = flytekit.current_context().logging

    fs = fsspec.filesystem("http")
    logger.info(f"Downloading {url} to {local_dest_path}")
    fs.get(url, local_dest_path)
    logger.info("File downloaded successfully!")

    return local_dest_path


def download_and_configure_vscode(version: str) -> str:
    """Download and configure vscode."""
    logger = flytekit.current_context().logging
    code_server_cli = which(CODE_SERVER_CLI_NAME)
    if code_server_cli is not None:
        logger.info(f"Code server binary already exists at {code_server_cli}")
        logger.info("Skipping downloading coe server")
        return code_server_cli

    # Download code-server
    logger.info(f"Code server is not in $PATH, downloading code server to {DOWNLOAD_DIR}...")
    DOWNLOAD_DIR.mkdir(exist_ok=True)

    machine_info = platform.machine()
    logger.info(f"machine type: {machine_info}")

    if machine_info == "aarch64":
        arch = "arm64"
    elif machine_info == "x86_64":
        arch = "amd64"
    else:
        msg = (
            "Automatic download is only supported on AMD64 and ARM64 architectures. "
            "If you are using a different architecture, please visit the code-server official "
            "website to manually download the appropriate version for your image."
        )
        raise ValueError(msg)

    code_server_tar_path_url = CODE_SERVER_DOWNLOAD_URL.format(version=version, arch=arch)
    code_server_local_path = DOWNLOAD_DIR / "code-server.tar.gz"
    download_http(code_server_tar_path_url, code_server_local_path)
    with tarfile.open(code_server_local_path, "r:gz") as tar:
        tar.extractall(path=DOWNLOAD_DIR)

    code_server_cli = (
        DOWNLOAD_DIR / CODE_SERVER_CLI_NAME_TEMPLATE.format(version=version, arch=arch) / "bin" / CODE_SERVER_CLI_NAME
    )
    return os.fspath(code_server_cli)


def monitor_vscode(child_process):
    logger = flytekit.current_context().logging
    start_time = time()

    while True:
        if not HEARTBEAT_PATH.exists():
            delta = time() - start_time
        else:
            delta = time() - HEARTBEAT_PATH.stat().st_mtime

        logger.info(f"The latest activity on code server was {delta} ago.")
        if delta > DEFAULT_UP_SECONDS:
            logger.info(f"code server is idle for more than {DEFAULT_UP_SECONDS} seconds. Terminating...")
            child_process.terminate()
            child_process.join()
            sys.exit(0)

        sleep(HEARTBEAT_CHECK_SECONDS)


def start_vscode_service(workspace_root: Path):
    code_server_cli = download_and_configure_vscode(CODE_SERVER_VERSION)

    host = f"0.0.0.0:{DEFAULT_PORT}"

    command = [
        code_server_cli,
        "--bind-addr",
        host,
        "--disable-workspace-trust",
        "--auth",
        "none",
    ]

    workspace_root = workspace_root.absolute()

    command.append(os.fspath(workspace_root))

    child_process = Process(target=run, args=[command], kwargs={"text": True})
    child_process.start()
    monitor_vscode(child_process)
